#!/usr/bin/env ruby
# encoding: utf-8

#--
#   Copyright 2014 Luke Spangler
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#++

# Library containing two classes which make
# TCP connections significantly easier.
# Eventually the other libraries will call
# this one for network communications.
#
#
# Author::				Luke Spangler
# Email::				m27frogy.roblox@gmail.com

require "abc"
require "accu-encrypt"

# The wrapper for TCPServer.
class ServerWrapper
	def initialize(port=2000)
		@port = TCPServer.open port
		catch :close do
			loop do
				@server = @port.accept
				@password,@mode = "password",:double
				yield self,@server if block_given?
				@server.close
			end
		end
		@port.close
	end
	
	# Listens for a generic network communication
	# and yields to a passed block (if any).
	def listen()
		msg = @server.gets.chomp
		yield msg if block_given?
		return msg
	end
	
	# Listens for multiple lines, appends them together
	# and stops at END.
	def listen_lines()
		lines = []
		while line = @server.gets.chomp and line != "END" do
			lines << line
		end
		lines
	end
		
	# Listens for a network code.
	def listen_code()
		listen { |msg|
			if msg.slice(0..1) == "C:" then
				return msg.slice(2..msg.length).to_i
			else
				return msg.slice(0..1)
			end
		}
	end
	
	# Listen for an encrypted connection.
	def listen_encrypted()
		if listen_code == 3 then
			password,source = EncryptDecrypt.decrypt(@password,((listen_lines).join),@mode)
			return source
		else
			raise "Sent code was not 3!"
		end
	end
		
	# Sends a plaintext message.
	def send(text)
		@server.puts text
	end
		
	# Sends multiple lines and appends
	# END to end the send.
	def send_lines(lines)
		array = lines.split("\n")
		array.each do |line|
			send(line)
		end
		send("END")
	end
		
	# Send over an encrypted multiline message.
	def send_encrypted(lines)
		secret,text = EncryptDecrypt.encrypt(@password,lines,@mode)
		send_code(3)
		send(text)
		send("END")
	end
		
	# Sends a code.
	def send_code(code)
		send("C:#{code}")
	end

	# Set encryption password.
	def set_password(string)
		warn "Argument 1 is not a string!  Continuing anyway." if not string.is_a? String
		@password = string.to_s
	end
	
	# Throws :close to end the serving process.
	def close
		throw :close
	end
end

#--------------------------------------------
#--------------------------------------------
#--------------------------------------------
#--------------------------------------------

# The wrapper for TCPClient.
class ClientWrapper
	def initialize(port=2000,ip="127.0.0.1")
		@client = TCPSocket.open(ip, port)
		@password,@mode = "password",:double
		yield self,@client if block_given?
		@client.close
	end
	
	# Listens for a generic network communication
	# and yields to a passed block (if any).
	def listen()
		msg = @client.gets.chomp
		yield msg if block_given?
		return msg
	end
	
	# Listens for multiple lines, appends them together
	# and stops at END.
	def listen_lines()
		lines = []
		while line = @client.gets.chomp and line != "END" do
			lines << line
		end
		lines
	end
		
	# Listens for a network code.
	def listen_code()
		listen { |msg|
			if msg.slice(0..1) == "C:" then
				return msg.slice(2..msg.length).to_i
			else
				return msg.slice(0..1)
			end
		}
	end
	
	# Listen for an encrypted connection.
	def listen_encrypted()
		if listen_code == 3 then
			password,source = EncryptDecrypt.decrypt(@password,((listen_lines).join),@mode)
			return source
		else
			raise "Sent code was not 3!"
		end
	end
		
	# Sends a plaintext message.
	def send(text)
		@client.puts text
	end
		
	# Sends multiple lines and appends
	# END to end the send.
	def send_lines(lines)
		array = lines.split("\n")
		array.each do |line|
			send(line)
		end
		send("END")
	end
		
	# Send over an encrypted multiline message.
	def send_encrypted(lines)
		secret,text = EncryptDecrypt.encrypt(@password,lines,@mode)
		send_code(3)
		send(text)
		send("END")
	end
		
	# Sends a code.
	def send_code(code)
		send("C:#{code}")
	end
	
	# Set encryption password.
	def set_password(string)
		warn "Argument 1 is not a string!  Continuing anyway." if not string.is_a? String
		@password = string.to_s
	end
end
