#--
#   Copyright 2014 Luke Spangler
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#++

# This is a library which uses some hacky references
# together with lots of math and loops to split
# a terminal into windows, via the module WindowTerminal.
# There are also several smaller classes declared
# which include:
#
# Orientation::			A class for orienting objects.
# Text::				A class for printing text in a window, based on an Orientation object.
# ColoredText::			A subclass of Text which adds compatibility for colored text.
# WrappedText::			A subclass of Text which allows text to be wrapped into a window.
# ColoredWrappedText::	A subclass of WrappedText which adds support for colored text.
# Line::				A class which creates a horizontal line across a window.
# Window::				The main class which handles the rendering of basic windows and their objects in a terminal emulator.
# WindowManager::		A class which allows simple(r) handling and rendering of windows and their objects.
#
#
# Author::				Luke Spangler
# Email::				m27frogy.roblox@gmail.com

require "highline/import"

# The primary module in-which the majority
# of classes and methods are bundled.
module WindowTerminal
	TIOCGNWINSZ = 0x5413
	STDOUT_HANDLE = 0xFFFFFFF5
	@@windows = []
	
	# A class which stores the Orientation
	# of an object.
	class Orientation
		attr_reader :x,:y
		
		# Initializes an Orienation object.
		def initialize(x=0,y=0)
			if x != -1 and x != 0 and x != 1 then
				x = 0
			end
			if y != -1 and y != 0 and y != 1 then
				y = 0
			end
			@x = x
			@y = y
		end
		
		# Makes the to_s return a more
		# "bootiful" string.
		def to_s
			"#{@x}|#{@y}"
		end
		
		# Allows for array-like indexing.
		# 0 - @x, 1 - @y
		def [](index)
			assert RangeError "Index must be either 0 or 1!" if not (index == 0 or index == 1)
			if index == 0 then
				return @x.dup
			else
				return @y.dup
			end
		end
	end
	
	# Aclass for use with the Window object
	# in rendering strings with Orientation.
	class Text
		attr_reader :orientation,:text, :y
		
		# Initializes a Text object.
		def initialize(orientation,text,y)
			raise "Orientation must be passed!" if not orientation.is_a? Orientation
			raise "String must be passed!" if not text.is_a? String
			@orientation = orientation
			@text = text
			@y = y
		end
		
		# Allows the @text instance variable
		# to be mutable.
		def set_text(text)
			@text = text
		end
		
		# Gets the length of the objects string
		# while ignoring the character 27.
		def get_length(text=@text)
			#--
			length = 0
			text.each_char {|char|
				if char.ord != 27  then
					length += 1
				end
			}
			#puts length
			return length
			#++
		end
		
		# Renders a line of text.
		def render_line(line,width,*args)
			#--
			line = line.dup
			if get_length() == @text.length and get_length.even? then
				@text << " "
			end
			if args[0] == nil then
				args[0] = 0
			end
			start = args[0]
			if @orientation.x == 0 then
				length = get_length()
				real_length = @text.length
				difference = (length - real_length).abs
				# Calculate centering.
				centered_width = (get_length() / 2).floor
				centered_length = ((width - 1) / 2).floor
				
				range = (centered_length-centered_width)..(centered_length+centered_width)
				# Add stretch space for colored text.
				if difference > 0 then
					char_existing = line[range.begin]
					append_string = ""
					#puts length,real_length
					(difference+(difference*2.5).ceil).times {|v|
						append_string << char_existing
					}
					#puts line[(range.begin)..(real_length -1)]
					line = line[0..range.begin] + append_string + line[(range.begin)..(width)]
					range = (range.begin+difference)..(range.end+difference)
				end
				line[range] = @text
				if get_length(line) > width then
					#puts "wut"
					# RAIG GLITCH AHX AHOGWHGOHWOGHWOGHGH
					line[get_length(line)-start-1] = ""
				end
			elsif @orientation.x == -1 then
				length = get_length()
				real_length = @text.length
				difference = (length - real_length).abs
				#puts difference
				range = (start)..(length+1)
				#puts range
				# Add stretch space for colored text.
				if difference > 0 then
					char_existing = line[range.begin]
					append_string = ""
					puts length,real_length
					(difference+(difference*2.5).ceil).times {|v|
						append_string << char_existing
					}
					#puts line[(range.begin)..(real_length -1)]
					puts "|" + line[0..range.begin-1] + "|"
					line = line[0..(range.begin-1)] + append_string + line[(range.begin)..(width)]
					#range = (range.begin+difference)..(range.end+difference)
				end
				line[range] = @text
			elsif @orientation.x == 1 then
				length = get_length()
				real_length = @text.length
				max_length = width - start - 1
				difference = (length - real_length).abs
				#puts difference
				range = (max_length - length + 1)..(max_length)
				#puts range
				# Add stretch space for colored text.
				if difference > 0 then
					char_existing = line[range.begin]
					append_string = ""
					puts length,real_length
					(difference+(difference*3).ceil).times {|v|
						append_string << char_existing
					}
					puts "|" + line[(range.end+1)..(range.end + start)] + "|"
					#puts (length/2.5).ceil
					line = line[0..(range.begin)] + append_string + line[(range.end-((length/2)-2).ceil)..(range.end + start)]
					range = (range.begin+difference*3.5)..(range.end)
				end
				line[range] = @text
			end
			return line
			#++
		end
		
		# Makes to_s return something
		# more meaningful.
		def to_s
			"<ColoredText @orientation=#{@orientation.to_s}, @y=#{@y} >"
		end
	end
	
	# A subclass of Text for colored window text.
	class ColoredText < Text
		#--
		#say(%{Here's some <%= color('dark red text', RED) %>.})
		#++
		
		# Initializes a ColoredText object.
		def initialize(orientation,text,y,color=:green)
			@color = color
			text = colorize(text)
			super(orientation,text,y)
		end
		
		def colorize(text)
			if WindowTerminal.os == :linux then
				if @color == :green then
					text = "\e[#{32}m#{text}\e[0m"
				elsif @color == :red then
					text = "\e[#{31}m#{text}\e[0m"
				elsif @color == :yellow then
					text = "\e[#{33}m#{text}\e[0m"
				elsif @color == :pink then
					text = "\e[#{35}m#{text}\e[0m"
				end
			end
			text
		end
		private :colorize
		
		# Sets text while taking into
		# account colorization.
		def set_text(text)
			super(colorize(text))
		end
		
		# Makes to_s return something
		# more meaningful.
		def to_s
			"<ColoredText @orientation=#{@orientation.to_s}, @y=#{@y}, @color=#{@color.to_s} >"
		end
	end
	
	# A subclass of Text which allows for
	# wrapped text within a Window.
	class WrappedText < Text
		undef :render_line
		
		# Initializes a WrappedText object.
		def initialize(orientation,text,y,mode=:destroy)
			@mode = mode
			super(orientation,text,y)
		end
		
		# Renders the WrappedText object
		# based on passed lines and
		# returns the modified lines.
		def render(lines,cols,rows,padding=2)
			#--
			lines = lines.dup
			words = wrap(@text,rows-padding*2)
			if words.length > (cols - @y) then
				words.slice!((cols - @y + 1)..(words.length-1))
			end
			range = (@y - 1)..(cols-1)
			sum = 0
			ending = cols - padding - 1
			lines.each_index { |index|
				if (range === index and words[sum]) and index < ending then
					#puts "|" + words[sum] + "|"
					lines[index] = WindowTerminal::Text.new(orientation,words[sum],0).render_line(lines[index],rows,padding)
					sum += 1
				end
			}
			return lines
			#++
		end
		
		private
		
		# Wraps a string based on width.
		#	<i>Credit: RubyCookbook</i>
		def wrap(s,width)
			#--
			if @mode == :destroy then
				s = s.dup
				s = s.split(/\s+/) if s.is_a? String
				lines = []
				line = ""
				#puts s.length
				s.each do |word|
					if line.size + word.size >= width then
						lines << line
						line = word
					elsif line.empty? then
						line = word
					else
						line << " " << word
					end
				end
				
				lines << line if line
				#puts lines.length
				return lines
			else
				string = s.gsub(/(.{1,#{width}})( |\Z)/, "\\1\n").split(/\n/)
				#puts string
				string
			end
		end
		#++
	end
	
	# A subclass of WrappedText which adds compatibility for
	# different colors.
	class ColoredWrappedText < WrappedText
		
		# Initializes a ColoredWrappedText object.
		def initialize(orientation,text,y,color=:green,mode=:destroy)
			@color = color
			if WindowTerminal.os == :linux then
				if @color == :green then
					text = "\e[#{32}m#{text}\e[0m"
				elsif @color == :red then
					text = "\e[#{31}m#{text}\e[0m"
				elsif @color == :yellow then
					text = "\e[#{33}m#{text}\e[0m"
				elsif @color == :pink then
					text = "\e[#{35}m#{text}\e[0m"
				end
			end
			super(orientation,text,y,mode)
		end
		
	end
	
	# A class which renders a simple
	# horizontal line on a Window.
	class Line
		attr_reader :y
		
		# Initializes a Line object.
		def initialize(y)
			@y = y
		end
		
		# Renders on the passed line
		# and returns the modified line.
		def render_line(line,width,*args)
			line.gsub(" ","-")
		end
	end
	
	# The main class which handles each individual
	# terminal window.
	class Window
		attr_reader :orientation,:objects
		
		# Initializes Window object.
		def initialize(orientation,name)
			raise "Orientation must be passed!" if not orientation.is_a? Orientation
			@orientation = orientation
			@name = name
			@objects = []
			add_object(WindowTerminal::ColoredText.new(WindowTerminal::Orientation.new,@name,0,:red))
			add_object(WindowTerminal::Line.new(1))
		end
		
		# Makes the to_s method return
		# something a bit more meaningful.
		def to_s
			"<Window: @orientation = #{@orientation.to_s}, @name = #{@name} >"
		end
		
		# Adds a single object to the Window's
		# object list.
		def add_object(object)
			raise ArgumentError "Argument must be a renderable object." if (not (object.respond_to?(:render_line) or object.respond_to?(:render)))
			@objects << object
		end
		
		# Adds all passed arguments to the
		# Window's object list.
		def add_objects(*items)
			raise ArgumentError "Argument missing" if items.length == 0
			items.each { |item|
				add_object(item)
			}
		end
		
		# Yields the passed block
		# with each object.
		def for_objects()
			@objects.each {|object|
				yield object
			}
		end
		
		# Sets the Window's current
		# name and changes the 
		# displayed name accordingly.
		def set_name(name)
			raise TypeError "Name must be an string!" if (not name.is_a? String)
			@name = name
			if @objects[0].respond_to? :set_text then
				@objects[0].set_text @name
			end
		end
		
		# Returns Window's current
		# name.
		def get_name()
			@name
		end
		
		# Includes the Window in the
		# WindowTerminal module screen_render.
		def include_in_render(bool=true)
			if bool and not WindowTerminal.get_windows().include? self then
				WindowTerminal.add_window self
			elsif not bool and WindowTerminal.get_windows().include? self then
				WindowTerminal.remove_window self
			end
		end
		
		# Allows for array-like indexing
		# of the Window's objects.
		def [](index)
			@objects[index]
		end
	end
	
	# Makes handling and switching between
	# Window objects simpler by creating
	# an API that allows for objects
	# to be assigned to pages,
	# which can, in turn,
	# be displayed, hidden, or deleted
	# at leisure.
	class WindowManager
		attr_reader :pages
		
		# Initializes a WindowManager object.
		def initialize()
			@pages = []
			@displayed_page = nil
		end
		
		# Creates a new page with the passed
		# arguments and returns the page
		# number.
		def new_page(*items)
			@pages << []
			items.each { |item|
				if item.is_a? Window then
					@pages[-1] << item
				end
			}
			return @pages.length
		end
		
		# Displays the page at the number
		# specified; defaults to one.
		def display_page(num=1)
			raise(ArgumentError, "Argument 1 must be a valid integer!") if (not (num.is_a? Fixnum))
			hide_page(@displayed_page)
			@pages[num-1].each { |item|
				WindowTerminal.add_window(item)
			}
			@displayed_page = num
		end
		
		# Alias for display_page.
		def show(num=1)
			display_page(num)
		end
		
		# Hides a page by a number,
		# defaults to doing nothing.
		def hide_page(num=nil)
			if num then
				raise(ArgumentError, "Argument 1 must be a valid integer!") if (not (num.is_a? Fixnum))
				if num == @displayed_page then
					@pages[num-1].each { |item|
						WindowTerminal.remove_window(item)
					}
					@displayed_page = nil
				end
			end
		end
		
		# An alias for hide_page which
		# always passes the current
		# displayed page number,
		# if it exists.
		def hide()
			if @displayed_page then
				hide_page(@displayed_page)
			end
		end
		
		# Delete a page by page number.
		def remove_page(num)
			raise(ArgumentError, "Argument 1 must be a valid integer!") if (not (num.is_a? Fixnum))
			hide_page(num)
			@pages.delete_at(num-1)
		end
		
		# Gets all Text and Text subclass
		# objects on the page given and
		# returns them all as an array.
		def get_page_text(num)
			raise(ArgumentError, "Argument 1 must be a valid integer!") if (not (num.is_a? Fixnum))
			array = []
			@pages[num-1].each {|window|
				array << window.objects.dup
				array[-1].each {|item|
					if (not (item.respond_to? :set_text)) then
						array[-1].delete(item)
					end
				}
			}
			return array
		end
		
		# An alias for the WindowTerminal method
		# screen_render.
		def render()
			WindowTerminal.screen_render
		end
	end
	
	# This module contains methods that generate
	# simple window setups (perhaps for use
	# with WindowManager.)
	module Templates
		# Returns an array containing the
		# window objects generated by
		# the template number passed.
		def self.generate(num)
			raise TypeError "Num must be an integer!" if (not num.is_a? Fixnum)
			if num == 1 then
				self.simple()
			end
		end
		
		# Creates a simple, one window
		# setup.
		#
		# Corresponds to generate(1)
		def self.simple()
			win = WindowTerminal::Window.new(WindowTerminal::Orientation.new,"Window")
			text = WindowTerminal::WrappedText.new(WindowTerminal::Orientation.new,"",8,:preserve)
			win.add_object text
			return [win]
		end
		
		# Calls generate() with
		# (argument 1) + 1.
		def [](index)
			raise TypeError "Num must be an integer!" if (not num.is_a? Fixnum)
			generate(index+1)
		end
	end
	
	# Gets current terminal size based
	# on the operating system and
	# returns the result.
	def self.terminal_size
		if self.os == :windows then
			m_GetStdHAndle = Win32Api.new("kernel32","GetStdHandle",["L"],"L")
			m_GetConsoleScreenBufferInfo = Win32Api.new("kernel32","GetConsoleScreenBufferInfo",["L","P"],"L")
			
			format = "SSSSSssssSS"
			buf = ([0] * format.size).pack(format)
			stdout_handle = m_GetStdHandle.call(STDOUT_HANDLE)
			
			m_GetConsoleScreenBufferInfo.call(stdout_handle, buf)
			(bufx, bufy, curx, cury, wattr, left, top, right, bottom, maxx, maxy) = buf.unpack(format)
			return bottom - top + 1, right - left + 1
		else
			rows, cols = 25, 80
			buf = [ 0, 0, 0, 0 ].pack("SSSS")
			if STDOUT.ioctl(TIOCGNWINSZ, buf) >= 0 then
				rows, cols, row_pixels, col_pixels = buf.unpack("SSSS")[0..1]
			end
			return rows,cols
		end
	end
	
	# Adds a window to the render.
	def self.add_window(win)
		@@windows << win
	end
	
	# Remoes a window from the render.
	def self.remove_window(win)
		@@windows.delete(win)
	end
	
	# Returns the windows currently
	# being rendered.
	def self.get_windows()
		@@windows
	end
	
	# Returns current operating system
	# based on the return value of
	# ENV["OS"].
	def self.os
		op = :blah
		if ENV["OS"] == nil then
			op = :linux
		else
			op = :windows
		end
		return op
	end
	
	# Renders the screen using
	# current window objects.
	def self.screen_render
		#--
		rows,cols = *WindowTerminal.terminal_size
		line1 = ""
		line2 = ""
		cols.times {|num|
			line1 << "#"
			line2 << " "
		}
		line2[0] = "#"
		line2[-1] = "#"
		line1 << "\n"
		line2 << "\n"
		lines = ""
		rows.times {|num|
			if num != 0 and num != rows - 1 then
				lines << line2
			else
				lines << line1
			end
		}
		if @@windows.length == 1 then
			window = @@windows[0]
			# Add lines for window.
			lines = lines.split("\n")
			lines.each_index { |index|
				#puts index
				if index > 1 and index < rows - 2 then
					lines[index][1] = "|"
					lines[index][-2] = "|"
				end
			}
			lines[1].gsub!(" ","-")
			#lines[3].gsub!(" ","-")
			lines[-2].gsub!(" ","-")
			# Render window objects.
			window.for_objects{ |object|
				if object.respond_to? :render_line then
					lines[object.y+2] = object.render_line(lines[object.y+2],cols,2)
				elsif object.respond_to? :render then
					lines = object.render(lines,rows,cols)
				end
			}
			lines = lines.join("\n")
		elsif @@windows.length == 2 then
			window1,window2 = @@windows[0],@@windows[1]
			if window1.orientation.x != window2.orientation.x then # Box-Box
				if window1.orientation.x > window2.orientation.x then
					left_window = window2
					right_window = window1
				else
					left_window = window1
					right_window = window2
				end
				lines = lines.split("\n")
				lines_left = []
				lines_right = []
				width = (cols - 1)
				left_dimension = (width / 2).ceil
				right_dimension = width - left_dimension
				lines.each_index {|index|
					#puts "iteration"
					lines_left[index] = lines[index].slice(0..left_dimension)
					lines_right[index] = lines[index].slice(right_dimension..width)
				}
				[[left_window,lines_left],[right_window,lines_right]].each { |array|
					window = array[0]
					current_lines = array[1]
					current_lines.each_index { |index|
						if index > 1 and index < rows - 2 then
							current_lines[index][0] = "#"
							current_lines[index][-1] = "#"
							current_lines[index][1] = "|"
							current_lines[index][-2] = "|"
						end
					}
					local_cols = current_lines[0].length
					current_lines[1].gsub!(" ","-")
					current_lines[-2].gsub!(" ","-")
					window.for_objects{ |object|
						if object.respond_to? :render_line then
							current_lines[object.y+2] = object.render_line(current_lines[object.y+2],local_cols,2)
						end
					}
				}
				lines = []
				lines_left.each_index { |index|
					lines[index] = lines_left[index] + lines_right[index]
				}
				lines_left = []
				lines_right = []
				lines = lines.join("\n")
			else # Box / Box
				if window1.orientation.y > window2.orientation.y then
					up_window = window2
					down_window = window1
				else
					up_window = window1
					down_window = window2
				end
				lines = lines.split("\n")
				up_lines = []
				down_lines = []
				height = (rows - 1)
				up_dimension = (height / 2).ceil
				down_dimension = height - up_dimension
				lines.each_index {|index|
					#puts "iteration"
					if index <= up_dimension then
						up_lines << lines[index]
					else
						down_lines << lines[index]
					end
				}
				[[up_window,up_lines],[down_window,down_lines]].each { |array|
					window = array[0]
					current_lines = array[1]
					current_lines[1].gsub!(" ","-")
					old = current_lines[-1].dup
					current_lines[-1].gsub!(" ","-")
					if current_lines[-1] == old then
						current_lines[-2].gsub!(" ","-")
					end
					current_lines.each_index { |index|
						if index == 0 or index == rows - 1 then
							current_lines[index] = line1.dup.gsub("\n","")
#						elsif index == 1 or index == rows - 2 then
#							current_lines[index].gsub!(" ","-")
						else
							#current_lines[index][0] = "#"
							#current_lines[index][-1] = "#"
							if current_lines[index][1] != "#" then
								current_lines[index][1] = "|"
								current_lines[index][-2] = "|"
							end
						end
					}
					window.for_objects{ |object|
						if object.respond_to? :render_line then
							current_lines[object.y+2] = object.render_line(current_lines[object.y+2],cols,2)
						end
					}
				}
				lines = []
				up_lines.each { |line|
					lines << line
				}
				up_lines = []
				down_lines.each {|line|
					lines << line
				}
				down_lines = []
				lines = lines.join("\n")
			end
		end
		print lines
		#++
	end
	
	# Gets a string from user using
	# the highline library ask()
	# set to no echo.
	def self.ask_quietly()
		string = ask("") { |q| q.echo = '' }
		yield(string) if block_given?
		self.screen_render
		return string
	end
	
	# Gets a single character from
	# the terminal emulator
	# in linux, otherwise it
	# returns an empty string.
	#
	# Also yields the character to
	# a passed block before rendering.
	def self.getchr()
		if self.os == :linux then
			string = ""
			begin
				system("stty raw -echo")
				string = STDIN.getc
			ensure
				system("stty -raw echo")
			end
			yield string if block_given?
			self.screen_render
			return string
		else
			""
		end
	end
	
	# Similar to ask_quietly but
	# renders the screen after
	# each character is entered.
	#
	# A block may also be passed
	# which will be executed
	# just before a render with the
	# current character and the full
	# string thus far.
	def self.getchrs()
		if self.os == :linux then
			full = ""
			string = " "
			string = self.getchr()
			while (string.ord != 13) do
				if string.ord == 127 and full.length > 0 then
					full.slice!(full.length - 1)
				else
					full << string.ord
				end
				yield(string,full) if block_given?
				self.screen_render
				string = self.getchr()
			end
			return full
		else
			""
		end
	end
end
